/*

  CS 599, Spring 2010, Physically Based Modeling for Interactive Simulation and Games
  USC/Viterbi/Computer Science
  Assignment 1 starter code

  Your name:
  <write your name here>

*/

#include "jello.h"
#include "showCube.h"
#include "input.h"
#include "physics.h"

// camera parameters
double Theta = pi / 6;
double Phi = pi / 6;
double R = 4;

// mouse control
int g_iMenuId;
int g_vMousePos[2];
int g_iLeftMouseButton,g_iMiddleMouseButton,g_iRightMouseButton;

// number of images saved to disk so far
int sprite=0;

// these variables control what is displayed on screen
int shear=0, bend=0, structural=1, pause=0, viewingMode=0, saveScreenToFile=0;

struct world jello;

int windowWidth, windowHeight;

int numTimesteps = 0;
GLuint textures[6];

/*
 * Loads skybox textures into memory, referred to by GLuint textures[].
 */
void loadTextures(){
	char *data;
	char *filenames[] = {"world/neg_x.rgb", "world/neg_y.rgb",
							"world/neg_z.rgb", "world/pos_x.rgb",
							"world/pos_y.rgb", "world/pos_z.rgb"};
	FILE *file;
	data = (char *)malloc(512*512*3);

	for(int i = 0; i < 6; i++){
		file = fopen(filenames[i], "rb");
		if(file == NULL)
			printf("Could not read file %s", filenames[i]);
		fread(data, 512*512*3, 1, file);
		fclose(file);

		glGenTextures(1, &textures[i]);
		glBindTexture(GL_TEXTURE_2D, textures[i]);
		glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
		gluBuild2DMipmaps( GL_TEXTURE_2D, 3, 512, 512, GL_RGB, GL_UNSIGNED_BYTE, data);
	}
    free( data );
}

void myinit()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(90.0,1.0,0.01,1000.0);

  // set background color to grey
  glClearColor(0.5, 0.5, 0.5, 0.0);

  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_POLYGON_SMOOTH);
  glEnable(GL_LINE_SMOOTH);

  loadTextures();
  return; 
}

/*
 * Draw the skybox. Temporarily disables lighting and depth masking.
 */
void showSkybox(double x, double y, double z){
	float size = 10.0f;
	glDepthMask(GL_FALSE);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glColor3f(1.0f, 1.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslated(x, y, z);

	//Front
	glBindTexture(GL_TEXTURE_2D, textures[0]);
	glBegin(GL_QUADS);
		glTexCoord2f(.999f, .999f);
		glVertex3f(-size, -size, -size);	
		glTexCoord2f(.999f, .001f);
		glVertex3f(-size, -size, size);
		glTexCoord2f(.001f, .001f);
		glVertex3f(size, -size, size);
		glTexCoord2f(.001f, .999f);
		glVertex3f(size, -size, -size);
	glEnd();

	//Bottom
	glBindTexture(GL_TEXTURE_2D, textures[1]);
	glBegin(GL_QUADS);		
		glTexCoord2f(.999f, .001f);
		glVertex3f(size, size, -size);
		glTexCoord2f(.001f, .001f);
		glVertex3f(-size, size, -size);
		glTexCoord2f(.001f, .999f);
		glVertex3f(-size, -size, -size);
		glTexCoord2f(.999f, .999f);
		glVertex3f(size, -size, -size);
	glEnd();

	//Right
	glBindTexture(GL_TEXTURE_2D, textures[2]);
	glBegin(GL_QUADS);	
		glTexCoord2f(.999f, .001f);
		glVertex3f(size, -size, size);
		glTexCoord2f(.001f, .001f);
		glVertex3f(size, size, size);
		glTexCoord2f(.001f, .999f);
		glVertex3f(size, size, -size);
		glTexCoord2f(.999f, .999f);
		glVertex3f(size, -size, -size);
	glEnd();

	//Back
	glBindTexture(GL_TEXTURE_2D, textures[3]);
	glBegin(GL_QUADS);
		glTexCoord2f(.001f, .001f);
		glVertex3f(-size, size, size);
		glTexCoord2f(.001f, .999f);
		glVertex3f(-size, size, -size);
		glTexCoord2f(.999f, .999f);
		glVertex3f(size, size, -size);
		glTexCoord2f(.999f, .001f);
		glVertex3f(size, size, size);
	glEnd();

	//Top
	glBindTexture(GL_TEXTURE_2D, textures[4]);
	glBegin(GL_QUADS);				
		glTexCoord2f(.001f, .999f);
		glVertex3f(-size, size, size);
		glTexCoord2f(.999f, .999f);
		glVertex3f(size, size, size);
		glTexCoord2f(.999f, .001f);
		glVertex3f(size, -size, size);
		glTexCoord2f(.001f, .001f);
		glVertex3f(-size, -size, size);
	glEnd();


	//Left
	glBindTexture(GL_TEXTURE_2D, textures[5]);
	glBegin(GL_QUADS);
		glTexCoord2f(.001f, .999f);
		glVertex3f(-size, -size, -size);
		glTexCoord2f(.999f, .999f);
		glVertex3f(-size, size, -size);
		glTexCoord2f(.999f, .001f);
		glVertex3f(-size, size, size);
		glTexCoord2f(.001f, .001f);
		glVertex3f(-size, -size, size);
	glEnd();

	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
	glDepthMask(GL_TRUE);
}

void display()
{
	double x,y,z;
	if(!(numTimesteps % jello.n)){
	  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	  glMatrixMode(GL_MODELVIEW);
	  glLoadIdentity();

	  // camera parameters are Phi, Theta, R
	  x = R*cos(Phi)*cos(Theta);
	  y = R*sin(Phi)*cos(Theta);
	  z = R*sin(Theta);
	  gluLookAt(x, y, z,
				0.0,0.0,0.0, 0.0,0.0,1.0);

	  //draw the skybox
	  showSkybox(x, y, z);

	  /* Lighting */
	  /* You are encouraged to change lighting parameters or make improvements/modifications
		 to the lighting model . 
		 This way, you will personalize your assignment and your assignment will stick out. 
	  */

	  // global ambient light
	  GLfloat aGa[] = { 0.0, 0.0, 0.0, 0.0 };
	  
	  // light 's ambient, diffuse, specular
	  GLfloat lKa0[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd0[] = { 1.0, 1.0, 1.0, 1.0 };
	  GLfloat lKs0[] = { 1.0, 1.0, 1.0, 1.0 };

	  GLfloat lKa1[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd1[] = { 1.0, 0.0, 0.0, 1.0 };
	  GLfloat lKs1[] = { 1.0, 0.0, 0.0, 1.0 };

	  GLfloat lKa2[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd2[] = { 1.0, 1.0, 0.0, 1.0 };
	  GLfloat lKs2[] = { 1.0, 1.0, 0.0, 1.0 };

	  GLfloat lKa3[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd3[] = { 0.0, 1.0, 1.0, 1.0 };
	  GLfloat lKs3[] = { 0.0, 1.0, 1.0, 1.0 };

	  GLfloat lKa4[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd4[] = { 0.0, 0.0, 1.0, 1.0 };
	  GLfloat lKs4[] = { 0.0, 0.0, 1.0, 1.0 };

	  GLfloat lKa5[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd5[] = { 1.0, 0.0, 1.0, 1.0 };
	  GLfloat lKs5[] = { 1.0, 0.0, 1.0, 1.0 };

	  GLfloat lKa6[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd6[] = { 1.0, 1.0, 1.0, 1.0 };
	  GLfloat lKs6[] = { 1.0, 1.0, 1.0, 1.0 };

	  GLfloat lKa7[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat lKd7[] = { 0.0, 1.0, 1.0, 1.0 };
	  GLfloat lKs7[] = { 0.0, 1.0, 1.0, 1.0 };

	  // light positions and directions
	  GLfloat lP0[] = { -1.999, -1.999, -1.999, 1.0 };
	  GLfloat lP1[] = { 1.999, -1.999, -1.999, 1.0 };
	  GLfloat lP2[] = { 1.999, 1.999, -1.999, 1.0 };
	  GLfloat lP3[] = { -1.999, 1.999, -1.999, 1.0 };
	  GLfloat lP4[] = { -1.999, -1.999, 1.999, 1.0 };
	  GLfloat lP5[] = { 1.999, -1.999, 1.999, 1.0 };
	  GLfloat lP6[] = { 1.999, 1.999, 1.999, 1.0 };
	  GLfloat lP7[] = { -1.999, 1.999, 1.999, 1.0 };
	  
	  // jelly material color

	  GLfloat mKa[] = { 0.0, 0.0, 0.0, 1.0 };
	  GLfloat mKd[] = { 0.3, 0.3, 0.3, 1.0 };
	  GLfloat mKs[] = { 1.0, 1.0, 1.0, 1.0 };
	  GLfloat mKe[] = { 0.0, 0.0, 0.0, 1.0 };

	  /* set up lighting */
	  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, aGa);
	  glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	  glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

	  // set up cube color
	  glMaterialfv(GL_FRONT, GL_AMBIENT, mKa);
	  glMaterialfv(GL_FRONT, GL_DIFFUSE, mKd);
	  glMaterialfv(GL_FRONT, GL_SPECULAR, mKs);
	  glMaterialfv(GL_FRONT, GL_EMISSION, mKe);
	  glMaterialf(GL_FRONT, GL_SHININESS, 120);
	    
	  // macro to set up light i
	  #define LIGHTSETUP(i)\
	  glLightfv(GL_LIGHT##i, GL_POSITION, lP##i);\
	  glLightfv(GL_LIGHT##i, GL_AMBIENT, lKa##i);\
	  glLightfv(GL_LIGHT##i, GL_DIFFUSE, lKd##i);\
	  glLightfv(GL_LIGHT##i, GL_SPECULAR, lKs##i);\
	  glEnable(GL_LIGHT##i)
	  
	  LIGHTSETUP (0);
	  LIGHTSETUP (1);
	  LIGHTSETUP (2);
	  LIGHTSETUP (3);
	  LIGHTSETUP (4);
	  LIGHTSETUP (5);
	  LIGHTSETUP (6);
	  LIGHTSETUP (7);

	  // enable lighting
	  glEnable(GL_LIGHTING);    
	  glEnable(GL_DEPTH_TEST);

	  // show the cube
	  showCube(&jello);
	  glDisable(GL_LIGHTING);

	  // show the bounding box
	  showBoundingBox();
	 
	  glutSwapBuffers();
  }
}

void doIdle()
{
  char s[20]="picxxxx.ppm";
  int i;
  
  // save screen to file
  s[3] = 48 + (sprite / 1000);
  s[4] = 48 + (sprite % 1000) / 100;
  s[5] = 48 + (sprite % 100 ) / 10;
  s[6] = 48 + sprite % 10;

  if (saveScreenToFile==1)
  {
    saveScreenshot(windowWidth, windowHeight, s);
    //saveScreenToFile=0; // save only once, change this if you want continuos image generation (i.e. animation)
    sprite++;
  }

  if (sprite >= 300) // allow only 300 snapshots
  {
    exit(0);	
  }

  if (pause == 0)
  {
    // insert code which appropriately performs one step of the cube simulation:
	  if(!strcmp(jello.integrator, "RK4"))
		  RK4(&jello);
	  else if(!strcmp(jello.integrator, "Euler"))
		  Euler(&jello);
	  numTimesteps++;
  }

  glutPostRedisplay();
}

int main (int argc, char ** argv)
{
  if (argc<2)
  {  
    printf ("Oops! You didn't say the jello world file!\n");
    printf ("Usage: %s [worldfile]\n", argv[0]);
    exit(0);
  }

  readWorld(argv[1],&jello);

  glutInit(&argc,argv);
  
  /* double buffered window, use depth testing, 640x480 */
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  
  windowWidth = 640;
  windowHeight = 480;
  glutInitWindowSize (windowWidth, windowHeight);
  glutInitWindowPosition (0,0);
  glutCreateWindow ("Jello cube");

  /* tells glut to use a particular display function to redraw */
  glutDisplayFunc(display);

  /* replace with any animate code */
  glutIdleFunc(doIdle);

  /* callback for mouse drags */
  glutMotionFunc(mouseMotionDrag);

  /* callback for mouse movement */
  glutPassiveMotionFunc(mouseMotion);

  /* callback for mouse button changes */
  glutMouseFunc(mouseButton);

  /* register for keyboard events */
  glutKeyboardFunc(keyboardFunc);

  /* do initialization */
  myinit();

  /* forever sink in the black hole */
  glutMainLoop();

  return(0);
}

