/*

  CS 599, Spring 2010, Physically Based Modeling for Interactive Simulation and Games
  USC/Viterbi/Computer Science
  Assignment 1 starter code

*/

#include "jello.h"
#include "physics.h"

#define REST_STRUCTURAL	1.0/7.0
#define REST_SHEAR1		1.0/7.0 * sqrt(2.0)
#define REST_SHEAR2		1.0/7.0 * sqrt(3.0)
#define REST_BEND		2.0/7.0

/*
 * Computes the force of a single spring, given a displacement vector (l),
 * a rest length (r), and spring stiffness (kHook).
 */
void computeFHook(point &l, point &force, double r, double kHook){
	point unitl;
	double magl, temp;
				
	magl = sqrt(l.x*l.x + l.y*l.y + l.z*l.z);
	unitl.x = l.x/magl;
	unitl.y = l.y/magl;
	unitl.z = l.z/magl;
	temp = kHook*(magl-r);
	force.x += temp*unitl.x;
	force.y += temp*unitl.y;
	force.z += temp*unitl.z;
}

/*
 * Computes the acceleration for a single mass-point. i,j,k are the indices in the 
 * block of mass-points. Calculates F_kHook, F_Damping, F_Forcefield, and collision detection.
 */
void computeOneAcceleration(struct world *jello, struct point &out, int i, int j, int k){
	int x,y,z;
	int numZeros = 0, collisionkHook = 1000;
	point l, unitl, deltav, min, force = {0,0,0};
	double r, magl, temp, alpha, beta, gamma, fieldSize, offset;

	// F_kHook for all incident springs
	for(x = -1; x <= 1; x++)
		for(y = -1; y <= 1; y++)
			for(z = -1; z <= 1; z++){
				if(i+x>=0 && i+x<8 && j+y>=0 && j+y<8 && k+z>=0 && k+z<8 &&
					(x!=0 || y!=0 || z!=0)){
					numZeros = 0;
					if(!x) numZeros++;
					if(!y) numZeros++;
					if(!z) numZeros++;
					switch (numZeros){
						case 0:
							r = REST_SHEAR2;
							break;
						case 1:
							r = REST_SHEAR1;
							break;
						case 2:
							r = REST_STRUCTURAL;
							break;
					}
					l.x = jello->p[i+x][j+y][k+z].x - jello->p[i][j][k].x;
					l.y = jello->p[i+x][j+y][k+z].y - jello->p[i][j][k].y;
					l.z = jello->p[i+x][j+y][k+z].z - jello->p[i][j][k].z;
					computeFHook(l, force, r, jello->kElastic);

					if(numZeros == 2 && i+x*2>=0 && i+x*2<8 && j+y*2>=0 && j+y*2<8 && k+z*2>=0 && k+z*2<8){
						r = REST_BEND;
						l.x = jello->p[i+x*2][j+y*2][k+z*2].x - jello->p[i][j][k].x;
						l.y = jello->p[i+x*2][j+y*2][k+z*2].y - jello->p[i][j][k].y;
						l.z = jello->p[i+x*2][j+y*2][k+z*2].z - jello->p[i][j][k].z;
						computeFHook(l, force, r, jello->kElastic);
						l.x = jello->p[i+x][j+y][k+z].x - jello->p[i][j][k].x;
						l.y = jello->p[i+x][j+y][k+z].y - jello->p[i][j][k].y;
						l.z = jello->p[i+x][j+y][k+z].z - jello->p[i][j][k].z;
					}

					// F_Damping for all springs
					deltav.x = jello->v[i+x][j+y][k+z].x - jello->v[i][j][k].x;
					deltav.y = jello->v[i+x][j+y][k+z].y - jello->v[i][j][k].y;
					deltav.z = jello->v[i+x][j+y][k+z].z - jello->v[i][j][k].z;
					magl = sqrt(l.x*l.x + l.y*l.y + l.z*l.z);
					unitl.x = l.x / magl;
					unitl.y = l.y / magl;
					unitl.z = l.z / magl;
					temp = jello->dElastic * (deltav.x*l.x + deltav.y*l.y + deltav.z*l.z) / magl;
					force.x += temp * unitl.x;
					force.y += temp * unitl.y;
					force.z += temp * unitl.z;

					// F_forcefield
					if(jello->resolution != 0 && jello->resolution != 1 && 
						jello->p[i][j][k].x > -2 && jello->p[i][j][k].x < 2 &&
						jello->p[i][j][k].y > -2 && jello->p[i][j][k].y < 2 &&
						jello->p[i][j][k].z > -2 && jello->p[i][j][k].z < 2){
						fieldSize = 4.0/(double)(jello->resolution-1);
						offset = 0;
						if(jello->resolution % 2 == 0)
							offset = fieldSize/2.0;
						alpha = fmod(abs(jello->p[i][j][k].x) + offset, fieldSize);
						beta = fmod(abs(jello->p[i][j][k].y) + offset, fieldSize);
						gamma = fmod(abs(jello->p[i][j][k].z) + offset, fieldSize);
						min.x = (int)((jello->p[i][j][k].x + 2.0) * (jello->resolution-1) / 4.0);
						min.y = (int)((jello->p[i][j][k].y + 2.0) * (jello->resolution-1) / 4.0);
						min.z = (int)((jello->p[i][j][k].z + 2.0) * (jello->resolution-1) / 4.0);
						force.x += jello->forceField[(int)min.x * jello->resolution * jello->resolution + (int)min.y * jello->resolution + (int)min.z].x;
						force.y += jello->forceField[(int)min.x * jello->resolution * jello->resolution + (int)min.y * jello->resolution + (int)min.z].y;
						force.z += jello->forceField[(int)min.x * jello->resolution * jello->resolution + (int)min.y * jello->resolution + (int)min.z].z;
						/*for(int d = 0; d < 2; d++)
							for(int e = 0; e < 2; e++)
								for(int f = 0; f < 2; f++){
									temp = abs((1-d)-alpha) * abs((1-e)-beta) * abs((1-f)-gamma);
									force.x += jello->forceField[(int)(min.x+d) * (int)(min.y+e) + (int)(min.z+f)].x * temp;
									force.y += jello->forceField[(int)(min.x+d) * (int)(min.y+e) + (int)(min.z+f)].y * temp;
									force.z += jello->forceField[(int)(min.x+d) * (int)(min.y+e) + (int)(min.z+f)].z * temp;
								}*/
					}

						//Collision Detection
						if(jello->p[i][j][k].x < -2){
							min.x = -(jello->p[i][j][k].x+2);
							min.y = 0;
							min.z = 0;
							computeFHook(min, force, 0, collisionkHook);
						}
						if(jello->p[i][j][k].x > 2){
							min.x = -(jello->p[i][j][k].x-2);
							min.y = 0;
							min.z = 0;
							computeFHook(min, force, 0, collisionkHook);
						}
						if(jello->p[i][j][k].y < -2){
							min.x = 0;
							min.y = -(jello->p[i][j][k].y+2);
							min.z = 0;
							computeFHook(min, force, 0, collisionkHook);
						}
						if(jello->p[i][j][k].y > 2){
							min.x = 0;
							min.y = -(jello->p[i][j][k].y-2);
							min.z = 0;
							computeFHook(min, force, 0, collisionkHook);
						}
						if(jello->p[i][j][k].z < -2){
							min.x = 0;
							min.y = 0;
							min.z = -(jello->p[i][j][k].z+2);
							computeFHook(min, force, 0, collisionkHook);
						}
						if(jello->p[i][j][k].z > 2){
							min.x = 0;
							min.y = 0;
							min.z = -(jello->p[i][j][k].z-2);
							computeFHook(min, force, 0, collisionkHook);
						}
				}
			}
				out.x = force.x / jello->mass;
				out.y = force.y / jello->mass;
				out.z = force.z / jello->mass;
}

/* Computes acceleration to every control point of the jello cube, 
   which is in state given by 'jello'.
   Returns result in array 'a'. */
void computeAcceleration(struct world * jello, struct point a[8][8][8])
{
  /* for you to implement ... */
	int i,j,k;
	for(i = 0; i < 8; i++)
		for(j = 0; j < 8; j++)
			for(k = 0; k < 8; k++)
				computeOneAcceleration(jello, a[i][j][k], i,j,k);
}

/* performs one step of Euler Integration */
/* as a result, updates the jello structure */
void Euler(struct world * jello)
{
  int i,j,k;
  point a[8][8][8];

  computeAcceleration(jello, a);
  
  for (i=0; i<=7; i++)
    for (j=0; j<=7; j++)
      for (k=0; k<=7; k++)
      {
        jello->p[i][j][k].x += jello->dt * jello->v[i][j][k].x;
        jello->p[i][j][k].y += jello->dt * jello->v[i][j][k].y;
        jello->p[i][j][k].z += jello->dt * jello->v[i][j][k].z;
        jello->v[i][j][k].x += jello->dt * a[i][j][k].x;
        jello->v[i][j][k].y += jello->dt * a[i][j][k].y;
        jello->v[i][j][k].z += jello->dt * a[i][j][k].z;

      }
}

/* performs one step of RK4 Integration */
/* as a result, updates the jello structure */
void RK4(struct world * jello)
{
  point F1p[8][8][8], F1v[8][8][8], 
        F2p[8][8][8], F2v[8][8][8],
        F3p[8][8][8], F3v[8][8][8],
        F4p[8][8][8], F4v[8][8][8];

  point a[8][8][8];


  struct world buffer;

  int i,j,k;

  buffer = *jello; // make a copy of jello

  computeAcceleration(jello, a);

  for (i=0; i<=7; i++)
    for (j=0; j<=7; j++)
      for (k=0; k<=7; k++)
      {
         pMULTIPLY(jello->v[i][j][k],jello->dt,F1p[i][j][k]);
         pMULTIPLY(a[i][j][k],jello->dt,F1v[i][j][k]);
         pMULTIPLY(F1p[i][j][k],0.5,buffer.p[i][j][k]);
         pMULTIPLY(F1v[i][j][k],0.5,buffer.v[i][j][k]);
         pSUM(jello->p[i][j][k],buffer.p[i][j][k],buffer.p[i][j][k]);
         pSUM(jello->v[i][j][k],buffer.v[i][j][k],buffer.v[i][j][k]);
      }

  computeAcceleration(&buffer, a);

  for (i=0; i<=7; i++)
    for (j=0; j<=7; j++)
      for (k=0; k<=7; k++)
      {
         // F2p = dt * buffer.v;
         pMULTIPLY(buffer.v[i][j][k],jello->dt,F2p[i][j][k]);
         // F2v = dt * a(buffer.p,buffer.v);     
         pMULTIPLY(a[i][j][k],jello->dt,F2v[i][j][k]);
         pMULTIPLY(F2p[i][j][k],0.5,buffer.p[i][j][k]);
         pMULTIPLY(F2v[i][j][k],0.5,buffer.v[i][j][k]);
         pSUM(jello->p[i][j][k],buffer.p[i][j][k],buffer.p[i][j][k]);
         pSUM(jello->v[i][j][k],buffer.v[i][j][k],buffer.v[i][j][k]);
      }

  computeAcceleration(&buffer, a);

  for (i=0; i<=7; i++)
    for (j=0; j<=7; j++)
      for (k=0; k<=7; k++)
      {
         // F3p = dt * buffer.v;
         pMULTIPLY(buffer.v[i][j][k],jello->dt,F3p[i][j][k]);
         // F3v = dt * a(buffer.p,buffer.v);     
         pMULTIPLY(a[i][j][k],jello->dt,F3v[i][j][k]);
         pMULTIPLY(F3p[i][j][k],0.5,buffer.p[i][j][k]);
         pMULTIPLY(F3v[i][j][k],0.5,buffer.v[i][j][k]);
         pSUM(jello->p[i][j][k],buffer.p[i][j][k],buffer.p[i][j][k]);
         pSUM(jello->v[i][j][k],buffer.v[i][j][k],buffer.v[i][j][k]);
      }
         
  computeAcceleration(&buffer, a);


  for (i=0; i<=7; i++)
    for (j=0; j<=7; j++)
      for (k=0; k<=7; k++)
      {
         // F3p = dt * buffer.v;
         pMULTIPLY(buffer.v[i][j][k],jello->dt,F4p[i][j][k]);
         // F3v = dt * a(buffer.p,buffer.v);     
         pMULTIPLY(a[i][j][k],jello->dt,F4v[i][j][k]);

         pMULTIPLY(F2p[i][j][k],2,buffer.p[i][j][k]);
         pMULTIPLY(F3p[i][j][k],2,buffer.v[i][j][k]);
         pSUM(buffer.p[i][j][k],buffer.v[i][j][k],buffer.p[i][j][k]);
         pSUM(buffer.p[i][j][k],F1p[i][j][k],buffer.p[i][j][k]);
         pSUM(buffer.p[i][j][k],F4p[i][j][k],buffer.p[i][j][k]);
         pMULTIPLY(buffer.p[i][j][k],1.0 / 6,buffer.p[i][j][k]);
         pSUM(buffer.p[i][j][k],jello->p[i][j][k],jello->p[i][j][k]);

         pMULTIPLY(F2v[i][j][k],2,buffer.p[i][j][k]);
         pMULTIPLY(F3v[i][j][k],2,buffer.v[i][j][k]);
         pSUM(buffer.p[i][j][k],buffer.v[i][j][k],buffer.p[i][j][k]);
         pSUM(buffer.p[i][j][k],F1v[i][j][k],buffer.p[i][j][k]);
         pSUM(buffer.p[i][j][k],F4v[i][j][k],buffer.p[i][j][k]);
         pMULTIPLY(buffer.p[i][j][k],1.0 / 6,buffer.p[i][j][k]);
         pSUM(buffer.p[i][j][k],jello->v[i][j][k],jello->v[i][j][k]);
      }

  return;  
}
